# (c) 2019 Florian Franzen <Florian.Franzen@gmail.com >

from imageio.core import Format

from os import SEEK_SET
import numpy as np

from construct import this, Struct, Bytes, Const, If, Int64ul, Int32ul, Float64l, Byte


GreenbergString = Struct(
    "length" / Int32ul,  # > 1 < 1e3
    "data" / Bytes(this.length),
    Const(0, Byte)
)

CamCommandoHeader = Struct(
    "header_size" / Int32ul,  # < 1e5
    "camera_type" / GreenbergString,  # b'basler', b'aptina'
    "header_version" / Float64l,  # > 0 < 100 > 0.13
    "image_type" / GreenbergString,

    "bytes_per_pixel" / Int32ul,
    "bits_per_pixel" / Int32ul,
    "frame_bytes_on_disk" / Int32ul,

    "width" / Int32ul,
    "height" / Int32ul,
    "frame_rate" / Float64l,

    "packed" / If(this.header_version >= 0.12, Byte),

    "frame_count" / Int32ul,

    "sensor" / If(this.header_version >= 0.12, Struct(
        "offset" / Int32ul[2],
        "size" / Int32ul[2],
        "clock" / Int64ul,
        "exposure" / Float64l,
        "gain" / Float64l
    ))
    # ToDo: Check that we are not exceeding header size here!
)


class CamCommandoFormat(Format):
    """ Adds support to read ccv files with the imagio library """

    def _can_read(self, request):
        # The request object has:
        # request.filename: a representation of the source (only for reporting)
        # request.firstbytes: the first 256 bytes of the file.
        # request.mode[0]: read or write mode
        # request.mode[1]: what kind of data the user expects: one of 'iIvV?'

        try:
            # Check for CCV header
            header = CamCommandoHeader.parse(request.firstbytes)
            # Newer file version not supported
            assert (header.header_version <= 0.13)
            # Pack files are not supported
            assert (not header.packed)
            # Other pixel sizes not supported
            assert (header.bits_per_pixel == 8)
            # ToDo: Check file size based on frame count and replace asserts
        except:
            return False

        return True

    class Reader(Format.Reader):
        def _open(self):
            # Parse header
            self.header = CamCommandoHeader.parse_stream(self.request.get_file())

            self.size = (self.header.width, self.header.height)

            self.meta = {
                "size": self.size,
                "fps":  self.header.frame_rate,
                "length": self.header.frame_count,
                "camera_type": self.header.camera_type.data,
                "image_type":  self.header.image_type.data}

            if self.header.header_version >= 0.12:
                self.meta["sensor"] = self.header.sensor

        def _get_length(self):
            # Return number of frames
            return self.header.frame_count

        def _get_data(self, index):
            # Check if requested index is in range
            if index >= self.header.frame_count:
                raise IndexError("Image index %i > %i".format(index, self.header.frame_count))

            # Seek to request frame in file
            offset = self.header.header_size + self.header.frame_bytes_on_disk * index
            self.request.get_file().seek(offset, SEEK_SET)

            # Read frame from file
            dimension = (self.header.height, self.header.width)
            frame = np.fromfile(self.request.get_file(), np.uint8, dimension[0] * dimension[1]).reshape(dimension)

            # Read in additional fields
            index = Int32ul.parse_stream(self.request.get_file())
            timestamp = Float64l.parse_stream(self.request.get_file())

            return frame, {"index": index, "timestamp": timestamp}

        def _get_meta_data(self, index):
            if index is None:
                return self.meta

            # Move to end of frame
            offset = self.header.header_size \
                   + self.header.frame_bytes_on_disk * index \
                   + self.header.height * self.header.width
            self.request.get_file().seek(offset, SEEK_SET)

            # Read in additional fields
            index = Int32ul.parse_stream(self.request.get_file())
            timestamp = Float64l.parse_stream(self.request.get_file())

            return {"index": index, "timestamp": timestamp}



# (c) 2021 Kay-Michael Voit <kay@voits.net>

import os,sys
import numpy as np
import imageio
from ccvtools import rawio
import random
import string
import mutagen
import json

def fill(full_times_file,ccv_file,ccv_times_file):
    ccv_out_file = ccv_file[0:-4]+'_filled'+ccv_file[-4:]

    ccv_times = np.loadtxt(ccv_times_file)
    full_times = np.loadtxt(full_times_file)

    assert ccv_times[0] > full_times[0] - 0.001
    doesnotexist = np.zeros(full_times.size,dtype=np.bool)
    ccv_times_idx = 0
    for i,t in enumerate(full_times):
        if len(ccv_times) <= ccv_times_idx:
            doesnotexist[i:] = True
            break
        if not np.abs(ccv_times[ccv_times_idx]-t) < 0.001:
            doesnotexist[i] = True
        else:
            ccv_times_idx = ccv_times_idx+1

    reader = imageio.get_reader(ccv_file)

    camera_type_len = reader.header.camera_type.length+5
    image_type_len = reader.header.image_type.length+5
    frcount_offset = int((8*32 + 64 +8)/8 + camera_type_len + image_type_len)

    blackframe = np.zeros(reader.header.frame_bytes_on_disk,dtype=np.uint8)

    with open(ccv_file,'rb') as sourcefile:
        with open(ccv_out_file,'wb') as targetfile:
            targetfile.write(sourcefile.read(frcount_offset))
            targetfile.write(np.uint32(len(full_times)))
            sourcefile.seek(frcount_offset+4,os.SEEK_SET)
            targetfile.write(sourcefile.read(reader.header.header_size - 4 - frcount_offset))
            
            for writeblack in tqdm(doesnotexist):
                if writeblack:
                    targetfile.write(blackframe)
                else:
                    targetfile.write(sourcefile.read(reader.header.frame_bytes_on_disk))


def truncate(ccv_file,idx_range,ccv_out_file=None):
    reader = imageio.get_reader(ccv_file)
    camera_type_len = reader.header.camera_type.length+5
    image_type_len = reader.header.image_type.length+5
    frcount_offset = int((8*32 + 64 +8)/8 + camera_type_len + image_type_len)

    if ccv_out_file is None:
        ccv_tmp_file = f'.{ccv_file[0:-4]}_{"".join(random.choices(string.ascii_uppercase + string.digits, k=5))}_truncate.tmp'
    else:
        ccv_tmp_file = f'.{ccv_out_file[0:-4]}_{"".join(random.choices(string.ascii_uppercase + string.digits, k=5))}_truncate.tmp'
        
    with open(ccv_file,'rb') as sourcefile:
        with open(ccv_tmp_file,'wb') as targetfile:
            # Copy header
            targetfile.write(sourcefile.read(reader.header.header_size))
            
            # Write frame number to header
            targetfile.seek(frcount_offset,os.SEEK_SET)
            targetfile.write(np.uint32(len(idx_range)))
            targetfile.seek(reader.header.header_size,os.SEEK_SET)
            
            # Copy desired frames
            prev_fr_idx = None
            for (i,fr_idx) in enumerate(idx_range):
                # Save first frame from iterator
                if i==0:
                    start_frame = fr_idx
                
                # Set point in case iterator is not consecutive
                if prev_fr_idx is None or not fr_idx-prev_fr_idx==1:
                    sourcefile.seek(reader.header.header_size+reader.header.frame_bytes_on_disk*fr_idx,os.SEEK_SET) 
                    
                targetfile.write(sourcefile.read(reader.header.frame_bytes_on_disk))
                prev_fr_idx = fr_idx
            
            # Save last frame from iterator
            end_frame = fr_idx
            
    # Rename tmp file to final file name
    if ccv_out_file is None:
        ccv_out_file = "{ccv_file[0:-4]}_truncated_{start_frame}-{end_frame}.ccv".format(ccv_file=ccv_file,start_frame=start_frame,end_frame=end_frame)
    os.rename(ccv_tmp_file,ccv_out_file)

    
def convert(ccv_file,video_file,idx_range,fps=25,codec="libx264",quality=10,min_contrast=0,max_contrast=None,out_type=np.uint8):
    reader = imageio.get_reader(ccv_file)
    writer = imageio.get_writer(video_file, fps=fps, codec=codec, quality=quality)

    prev_fr_idx=None
    indices = np.zeros(shape=len(idx_range),dtype=int)
    timestamps = np.zeros(shape=len(idx_range),dtype=np.float)
    for (i,fr_idx) in enumerate(idx_range):
        # Set point in case iterator is not consecutive
        if prev_fr_idx is None or not fr_idx-prev_fr_idx==1:
            im = reader.get_data(fr_idx)
        else:
            im = reader.get_next_data()
        
        meta = reader.get_meta_data(fr_idx);
        indices[i] = meta["index"];
        timestamps[i] = meta["timestamp"];
        
        # Get max value of movie if not specified
        if i==0 and max_contrast is None:
            max_contrast = np.iinfo(np.asarray(im).dtype).max
                
        # Adjust contrast / Reduce to 8 bit
        im = np.uint64(im)
        im = (im-min_contrast)*np.iinfo(out_type).max/(max_contrast-min_contrast)
        im[im<0] = 0
        im[im>max_contrast] = np.iinfo(out_type).max
        writer.append_data(out_type(im))
        prev_fr_idx = fr_idx

    writer.close()

    headerdict = dict(reader.header)
    del headerdict['_io']
    headerdict['camera_type'] = headerdict['camera_type'].data.decode("ascii")
    headerdict['image_type'] = headerdict['image_type'].data.decode("ascii")
    headerdict['sensor'] = {'offset': list(headerdict['sensor'].offset),
                            'size': list(headerdict['sensor'].size),
                            'clock': headerdict['sensor'].clock,
                            'exposure': headerdict['sensor'].exposure,
                            'gain': headerdict['sensor'].gain
                           }
    headerdict['indices'] = [int(val) for val in indices]
    print(timestamps)
    headerdict['timestamps'] = [float(val) for val in timestamps]
    print(headerdict['timestamps'])
    
    if len(headerdict['sensor']['offset'])==1: headerdict['sensor']['offset'].append(headerdict['sensor']['offset'][0])
    if len(headerdict['sensor']['size'])==1: headerdict['sensor']['size'].append(headerdict['sensor']['size'][0])
    
    with open(video_file, 'r+b') as file:
        media_file = mutagen.File(file, easy=True)
        media_file['description'] = json.dumps(headerdict)
        media_file.save(file)



import pathlib
from setuptools import find_packages, setup
# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="bbo-ccvtools",
    version="1.1.8",
    description="Work with BBO CCV files",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/bbo-lab/ccvtools",
    author="BBO-lab @ caesar",
    author_email="kay-michael.voit@caesar.de",
    license="BSD",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=['ccvtools','ccvtools.rawio'],
    include_package_data=True,
    install_requires=["construct", "numpy", "imageio", "imageio-ffmpeg", "pyjson", "mutagen"],
)
